# -*- coding: utf-8 -*-
"""
    falcon_caching.backends.simple
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    The simple cache backend.

    :copyright: (c) 2019 by Zoltan Fedor.
    :license: MIT, see LICENSE for more details.

    :copyright: (c) 2018 by Peter Justin.
    :copyright: (c) 2010 by Thadeus Burgess.
    :license: BSD, see LICENSE for more details.
"""
from time import time

from falcon_caching.backends.base import BaseCache

try:
    import cPickle as pickle
except ImportError:  # pragma: no cover
    import pickle  # type: ignore


class SimpleCache(BaseCache):
    """Simple memory cache for single process environments.  This class exists
    mainly for the development server and is not 100% thread safe.  It tries
    to use as many atomic operations as possible and no locks for simplicity
    but it could happen under heavy load that keys are added multiple times.

    :param threshold: the maximum number of items the cache stores before
                      it starts deleting some.
    :param default_timeout: the default timeout that is used if no timeout is
                            specified on :meth:`~BaseCache.set`. A timeout of
                            0 indicates that the cache never expires.
    :param ignore_errors: If set to ``True`` the :meth:`~BaseCache.delete_many`
                          method will ignore any errors that occured during the
                          deletion process. However, if it is set to ``False``
                          it will stop on the first error. Defaults to
                          ``False``.
    """

    def __init__(self, threshold=500, default_timeout=300, ignore_errors=False):
        super(SimpleCache, self).__init__(default_timeout)
        self._cache = {}
        self.clear = self._cache.clear
        self._threshold = threshold
        self.ignore_errors = ignore_errors

    def _prune(self):
        if len(self._cache) > self._threshold:
            now = time()
            toremove = []
            for idx, (key, (expires, _)) in enumerate(self._cache.items()):
                if (expires != 0 and expires <= now) or idx % 3 == 0:
                    toremove.append(key)
            for key in toremove:
                self._cache.pop(key, None)

    def _normalize_timeout(self, timeout):
        timeout = BaseCache._normalize_timeout(self, timeout)
        if timeout > 0:
            timeout = time() + timeout
        return timeout

    def get(self, key):
        try:
            expires, value = self._cache[key]
            if expires == 0 or expires > time():
                return pickle.loads(value)
        except (KeyError, pickle.PickleError):
            return None

    def set(self, key, value, timeout=None):
        expires = self._normalize_timeout(timeout)
        self._prune()
        self._cache[key] = (
            expires,
            pickle.dumps(value, pickle.HIGHEST_PROTOCOL),
        )
        return True

    def add(self, key, value, timeout=None):
        expires = self._normalize_timeout(timeout)
        self._prune()
        item = (expires, pickle.dumps(value, pickle.HIGHEST_PROTOCOL))
        if key in self._cache:
            return False
        self._cache.setdefault(key, item)
        return True

    def delete(self, key):
        return self._cache.pop(key, None) is not None

    def has(self, key):
        try:
            expires, value = self._cache[key]
            return expires == 0 or expires > time()
        except KeyError:
            return False



from inspect import isclass
from typing import Any, Dict, List, Optional

from falcon_caching.middleware import Middleware
from falcon_caching.options import CacheEvictionStrategy


class Cache:
    """ This is the central class for the caching

    You need to initialize this object to setup the attributes of the caching
    and then supply the object's middleware to the Falcon app.

    Args:
        config (dict of str: str): Cache config settings

    Attributes:
        cache (:obj:`BaseCache`): An initialized 'CACHE_TYPE' cache from the backends.
        cache_args (list of str): Optional list passed during the cache class instantiation.
        cache_options (dict of str: str): Optional dictionary passed during the cache class instantiation.
        config (dict of str: str): Cache config settings
    """

    def __init__(self, config: Dict[str, Any]) -> None:
        """ Import and initialize the cache object for the requested 'CACHE_TYPE'
        and set its arguments and options and then store it on self.cache """

        # set the defaults for the config
        config.setdefault("CACHE_DEFAULT_TIMEOUT", 300)
        config.setdefault("CACHE_IGNORE_ERRORS", False)
        config.setdefault("CACHE_THRESHOLD", 500)
        config.setdefault("CACHE_KEY_PREFIX", "flask_cache_")
        config.setdefault("CACHE_MEMCACHED_SERVERS", None)
        config.setdefault("CACHE_MEMCACHED_USERNAME", None)
        config.setdefault("CACHE_MEMCACHED_PASSWORD", None)
        config.setdefault("CACHE_DIR", None)
        config.setdefault("CACHE_OPTIONS", None)
        config.setdefault("CACHE_ARGS", [])
        config.setdefault("CACHE_TYPE", "null")
        config.setdefault("CACHE_NO_NULL_WARNING", False)
        config.setdefault("CACHE_EVICTION_STRATEGY", CacheEvictionStrategy.time_based)

        self.config = config

        import_me = config["CACHE_TYPE"]
        from . import backends
        try:
            cache_obj = getattr(backends, import_me)
        except AttributeError:
            raise ImportError(f"{import_me} is not a valid Falcon-Caching backend")

        self.cache_args = config["CACHE_ARGS"][:]
        self.cache_options = {"default_timeout": config["CACHE_DEFAULT_TIMEOUT"]}

        if config["CACHE_OPTIONS"]:
            self.cache_options.update(config["CACHE_OPTIONS"])

        # initialize the cache_object
        self.cache = cache_obj(self.config, self.cache_args, self.cache_options)

    @property
    def middleware(self) -> 'Middleware':
        """ Falcon middleware integration
        """
        return Middleware(self.cache, self.config)

    @staticmethod
    def cached(timeout: int):
        """ This is the decorator used to decorate a resource class or the requested
        method of the resource class
        """
        def wrap1(class_or_method, *args):
            # is this about decorating a class or a given method?
            if isclass(class_or_method):
                # get all methods of the class that needs to be decorated (eg start with "on_"):
                for attr in dir(class_or_method):
                    if callable(getattr(class_or_method, attr)) and attr.startswith("on_"):
                        # decorate the given method:
                        setattr(class_or_method, attr, wrap1(getattr(class_or_method, attr)))

                return class_or_method
            else:  # this is to decorate the individual method
                class_or_method.to_be_cached = True

                def cache_wrap(cls, req, resp, *args, **kwargs):
                    class_or_method(cls, req, resp, *args, **kwargs)
                    req.context['cache'] = True
                    req.context['cache_timeout'] = timeout

                return cache_wrap

        return wrap1

    def has(self, *args, **kwargs) -> bool:
        """It determines if the given key is in the cache."""
        _h = self.cache.has(*args, **kwargs)
        # the Redis backend returns 0 for not existing keys
        if isinstance(_h, bool):
            return _h
        elif isinstance(_h, int):
            return False if _h == 0 else True
        else:
            raise ValueError(f"has() returned an unknown value '{_h}'")

    def get(self, *args, **kwargs) -> Any:
        """It returns the value for the given key from the cache."""
        return self.cache.get(*args, **kwargs)

    def set(self, *args, **kwargs) -> bool:
        """It stores the given key and value in the cache."""
        return self.cache.set(*args, **kwargs)

    def add(self, *args, **kwargs) -> bool:
        """It adds a given key and value to the cache, but only
        if no record which such key already exists.
        """
        return self.cache.add(*args, **kwargs)

    def delete(self, *args, **kwargs) -> bool:
        """It deletes the cached record based on the provided key."""
        return self.cache.delete(*args, **kwargs)

    def delete_many(self, *args, **kwargs) -> bool:
        """It deletes all cached record matching the list of keys provided."""
        return self.cache.delete_many(*args, **kwargs)

    def clear(self) -> bool:
        """It clears all cache - if the `CACHE_KEY_PREFIX` config attribute
        is used then it only removes key starting with that prefix, otherwise
        it flushes the whole database."""
        return self.cache.clear()

    def get_many(self, *args, **kwargs) -> List[Any]:
        """It returns the list of values matching the list of keys."""
        return self.cache.get_many(*args, **kwargs)

    def set_many(self, *args, **kwargs) -> bool:
        """It stores multiple records based on the dictionary of keys
        and values provided."""
        return self.cache.set_many(*args, **kwargs)

    def get_dict(self, *args, **kwargs) -> Dict[Any, Any]:
        """It returns the keys and values as dictionary for all requested keys."""
        return self.cache.get_dict(*args, **kwargs)

    def inc(self, *args, **kwargs) -> Optional[int]:
        """It increments and returns the value of a numerical cache record.
        Only works for Redis and Redis Sentinel!
        """
        return self.cache.inc(*args, **kwargs)

    def dec(self, *args, **kwargs) -> Optional[int]:
        """It decrements and returns the value of a numerical cache record.
        Only works for Redis and Redis Sentinel!
        """
        return self.cache.dec(*args, **kwargs)



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from falcon import API
    from falcon_caching.cache import Cache


def get_cache(app: 'API') -> 'Cache':
    """ Get the cache object from the app """
    return app._middleware[1][0].__self__


def get_cache_class(app: 'API'):
    """ Extracts the cache backend class from the Falcon app,
    so we can use it to determine when to skip certain tests
    """
    return get_cache(app).cache.__class__


def get_cache_eviction_strategy(app: 'API'):
    """ Extracts the cache eviction strategy from the Falcon app,
    so we can use it to determine when to skip certain tests
    """
    return get_cache(app).cache_config['CACHE_EVICTION_STRATEGY']


def delete_from_cache(app: 'API', path: str, method: str, request_body: str=None) -> None:
    """ Delete / remove a certain key from the cache
    """
    request_body = request_body if request_body else ''

    if path.endswith('/'):
        path = path[:-1]

    key = f'{path}:{method.upper()}:{request_body}'

    get_cache(app).cache.delete(key)



import errno
import json
import os
import pytest
import random
import shutil

from falcon import API, testing
from falcon_caching import Cache

try:
    __import__("pytest_xprocess")
    from xprocess import ProcessStarter
except ImportError:

    @pytest.fixture(scope="session")
    def xprocess():
        pytest.skip("pytest-xprocess not installed.")

# the different cache_types that will be tested
CACHE_TYPES = [
    'simple',
    'filesystem',
    'redis',
    'redissentinel',
    'uwsgi',
    'memcached',
    'gaememcached',
    'saslmemcached',
    'spreadsaslmemcached',
]

# what eviction strategies we will be testing
EVICTION_STRATEGIES = [
    'time-based',
    'rest-based',
    'rest-and-time-based'
]

# cache-busting methods
# eg which HTTP methods are busting the cache in 'rest-based'
# and 'rest-and-time-based' eviction strategy
CACHE_BUSTING_METHODS = [
    'POST',
    'PUT',
    'PATCH',
    'DELETE'
]

# we want to test the pruning, so we set the threshold low
# instead of the default 500
CACHE_THRESHOLD = 5

# which port the Redis server will be listening on
# which is started by xprocess
REDIS_PORT = 63799

# to test expiring cache, how long the cache expires
CACHE_EXPIRES = 1


# parametrized fixture to create caches with different types (eg backends)
@pytest.fixture(params=CACHE_TYPES)
def caches(request, tmp_path, redis_server, redis_sentinel_server, memcache_server):
    """ Time-based cache parametrized to generate a cache
    for each cache_type (eg backend)
    """
    if request.param == 'redissentinel' and os.getenv('TRAVIS', 'no') == 'yes':
        pytest.skip("Unfortunately on Travis Redis Sentinel currently can't be installed")

    # uwsgi tests should only run if running under uwsgi
    if request.param == 'uwsgi':
        try:
            import uwsgi
        except ImportError:
            pytest.skip("uWSGI could not be imported, are you running under uWSGI?")
            return None

    # build a dict of caches for each eviction strategy
    caches = {
        eviction_strategy:
            Cache(
                config={
                    'CACHE_EVICTION_STRATEGY': eviction_strategy,
                    'CACHE_TYPE': request.param,
                    'CACHE_THRESHOLD': CACHE_THRESHOLD,
                    'CACHE_DIR': tmp_path if request.param == 'filesystem' else None,
                    'CACHE_REDIS_PORT': REDIS_PORT
                }
            )
        for eviction_strategy in EVICTION_STRATEGIES
    }
    return caches


@pytest.fixture()
def cache_time_based(caches):
    """ Returns a cache instance, which can directly be used in
    the test_cache_backends.py to test the various cache backends
    """
    return caches['time-based'].cache


@pytest.fixture(scope="class")
def redis_server(xprocess):
    try:
        import redis
    except ImportError:
        pytest.skip("Python package 'redis' is not installed.")

    class Starter(ProcessStarter):
        pattern = "[Rr]eady to accept connections"
        args = ["redis-server", "--port", REDIS_PORT]

    try:
        xprocess.ensure("redis_server", Starter)
    except IOError as e:
        # xprocess raises FileNotFoundError
        if e.errno == errno.ENOENT:
            pytest.skip("Redis is not installed.")
        else:
            raise

    yield
    xprocess.getinfo("redis_server").terminate()


@pytest.fixture(scope="class")
def redis_sentinel_server(xprocess, redis_server):
    # on Travis there is no redis-sentinel, so we need to skip this,
    # but can't use pytest.skip() as that would bubble up to all tests
    if os.getenv('TRAVIS', 'no') == 'yes':
        yield
    else:
        try:
            import redis
        except ImportError:
            pytest.skip("Python package 'redis' is not installed.")

        # copy the sentinel_original.conf to sentinel.conf, as sentinel
        # will modify it
        shutil.copy("tests/sentinel_original.conf", "tests/sentinel.conf")

        class Starter(ProcessStarter):
            pattern = "monitor master mymaster"
            args = ["redis-sentinel", f"{os.getcwd()}/tests/sentinel.conf"]

        try:
            xprocess.ensure("redis_sentinel_server", Starter)
        except IOError as e:
            # xprocess raises FileNotFoundError
            if e.errno == errno.ENOENT:
                pytest.skip("Redis Sentinel is not installed.")
            else:
                raise

        yield
        xprocess.getinfo("redis_sentinel_server").terminate()


@pytest.fixture(scope="class")
def memcache_server(xprocess):
    try:
        import pylibmc as memcache
    except ImportError:
        try:
            from google.appengine.api import memcache
        except ImportError:
            try:
                import memcache
            except ImportError:
                pytest.skip(
                    "Python package for memcache is not installed. Need one of "
                    "pylibmc', 'google.appengine', or 'memcache'."
                )

    class Starter(ProcessStarter):
        pattern = ""
        args = ["memcached"]

    try:
        xprocess.ensure("memcached", Starter)
    except IOError as e:
        # xprocess raises FileNotFoundError
        if e.errno == errno.ENOENT:
            pytest.skip("Memcached is not installed.")
        else:
            raise

    yield
    xprocess.getinfo("memcached").terminate()


@pytest.fixture(params=EVICTION_STRATEGIES)
def app(request, caches):
    """ Creates a Falcon app with the given cache instance to be used
    by functional tests
    """
    # get the cache for the given eviction strategy
    cache = caches[request.param]

    class CachedResource:
        """ A cached resource with long expiration and all the different methods """
        @cache.cached(timeout=60)
        def on_get(self, req, resp):
            resp.body = json.dumps({'num': random.randrange(0, 100000)})

        @cache.cached(timeout=60)
        def on_post(self, req, resp):
            pass

        @cache.cached(timeout=60)
        def on_put(self, req, resp):
            pass

        @cache.cached(timeout=60)
        def on_patch(self, req, resp):
            pass

        @cache.cached(timeout=60)
        def on_delete(self, req, resp):
            pass

    class CachedResourceExpires:
        """ A cached resource with short expiration """
        @cache.cached(timeout=CACHE_EXPIRES)
        def on_get(self, req, resp):
            resp.body = json.dumps({'num': random.randrange(0, 100000)})

        @cache.cached(timeout=CACHE_EXPIRES)
        def on_post(self, req, resp):
            pass

    @cache.cached(timeout=CACHE_EXPIRES)
    class ClassCachedResourceExpires:
        """ A cached resource which cached on the class level """
        def on_get(self, req, resp):
            resp.body = json.dumps({'num': random.randrange(0, 100000)})

        def on_post(self, req, resp):
            pass

    app = API(middleware=cache.middleware)

    app.add_route('/randrange_cached', CachedResource())
    app.add_route('/randrange_cached_expires', CachedResourceExpires())
    app.add_route('/randrange_class_cached_expires', ClassCachedResourceExpires())

    return app


@pytest.fixture()
def client(app):
    """ Creates a Falcon test client
    """
    return testing.TestClient(app)



# -*- coding: utf-8 -*-
"""
    flask_caching.backends.uwsgicache
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    The uWSGI caching backend.

    :copyright: (c) 2019 by Zoltan Fedor.
    :license: MIT, see LICENSE for more details.

    :copyright: (c) 2018 by Peter Justin.
    :copyright: (c) 2010 by Thadeus Burgess.
    :license: BSD, see LICENSE for more details.
"""
import platform

from falcon_caching.backends.base import BaseCache

try:
    import cPickle as pickle
except ImportError:  # pragma: no cover
    import pickle


class UWSGICache(BaseCache):
    """ Implements the cache using uWSGI's caching framework.

    .. note::
        This class cannot be used when running under PyPy, because the uWSGI
        API implementation for PyPy is lacking the needed functionality.

    :param default_timeout: The default timeout in seconds.
    :param cache: The name of the caching instance to connect to, for
        example: mycache@localhost:3031, defaults to an empty string, which
        means uWSGI will cache in the local instance. If the cache is in the
        same instance as the werkzeug app, you only have to provide the name of
        the cache.
    """

    def __init__(self, default_timeout=300, cache=""):
        super(UWSGICache, self).__init__(default_timeout)

        if platform.python_implementation() == "PyPy":
            raise RuntimeError(
                "uWSGI caching does not work under PyPy, see "
                "the docs for more details."
            )

        try:
            import uwsgi

            self._uwsgi = uwsgi
        except ImportError:
            raise RuntimeError(
                "uWSGI could not be imported, are you " "running under uWSGI?"
            )

        self.cache = cache

    def get(self, key):
        rv = self._uwsgi.cache_get(key, self.cache)
        if rv is None:
            return
        return pickle.loads(rv)

    def delete(self, key):
        return self._uwsgi.cache_del(key, self.cache)

    def set(self, key, value, timeout=None):
        return self._uwsgi.cache_update(
            key,
            pickle.dumps(value),
            self._normalize_timeout(timeout),
            self.cache,
        )

    def add(self, key, value, timeout=None):
        return self._uwsgi.cache_set(
            key,
            pickle.dumps(value),
            self._normalize_timeout(timeout),
            self.cache,
        )

    def clear(self):
        return self._uwsgi.cache_clear(self.cache)

    def has(self, key):
        return self._uwsgi.cache_exists(key, self.cache) is not None



""" Test the cache backends using a raw Cache() object for each different
backed type.

Some of these tests were shamelessly taken from the Flask-Caching library -
just like the backends were.
"""
import os
import pytest
import random
import time

from tests.conftest import CACHE_THRESHOLD
from falcon_caching import Cache
from falcon_caching.backends.memcache import (
    MemcachedCache,
    SASLMemcachedCache,
    SpreadSASLMemcachedCache
)
from falcon_caching.backends.redis import Redis, RedisSentinel


def test_cache_set(cache_time_based):
    cache = cache_time_based
    cache.set("hi", "hello")

    # on Travis the below fails for SpreadSASLMemcachedCache
    # but locally it doesn't...
    if not isinstance(cache, SpreadSASLMemcachedCache)\
            or os.getenv('TRAVIS', 'no') == 'no':
        assert cache.has("hi") in [True, 1]
        assert cache.has("nosuchkey") in [False, 0]

    assert cache.get("hi") == "hello"


def test_cache_add(cache_time_based):
    cache = cache_time_based
    # on Travis the below fails for SpreadSASLMemcachedCache
    # but locally it doesn't...
    if isinstance(cache, SpreadSASLMemcachedCache)\
            and os.getenv('TRAVIS', 'no') == 'yes':
        pytest.skip("Skipping for SpreadSASLMemcachedCache on Travis")

    cache.add("hi", "hello")
    assert cache.get("hi") == "hello"

    cache.add("hi", "foobar")
    assert cache.get("hi") == "hello"


def test_cache_delete(cache_time_based):
    cache = cache_time_based
    cache.set("hi", "hello")
    cache.delete("hi")
    assert cache.get("hi") is None


def test_cache_delete_many(cache_time_based):
    cache = cache_time_based
    cache.set("hi", "hello")
    cache.set("hi2", "hello")
    cache.delete_many("hi", "hi2")
    assert cache.get("hi") is None
    assert cache.get("hi2") is None


def test_cache_unlink_if_not(cache_time_based):
    cache = cache_time_based
    # not every cache type has an unlink() method
    if hasattr(cache, 'unlink'):
        cache.set("biggerkey", "test" * 100)
        cache.unlink("biggerkey")
        assert cache.get("biggerkey") is None

        cache.set("biggerkey1", "test" * 100)
        cache.set("biggerkey2", "test" * 100)
        cache.unlink("biggerkey1", "biggerkey2")
        assert cache.get("biggerkey1") is None
        assert cache.get("biggerkey2") is None


def test_cache_delete_many_ignored():
    c = Cache(config={"CACHE_TYPE": "simple", "CACHE_IGNORE_ERRORS": True, "EVICTION_STRATEGY": 'time-based'})
    cache = c.cache

    cache.set("hi", "hello")
    assert cache.get("hi") == "hello"
    cache.delete_many("ho", "hi")
    assert cache.get("hi") is None


def test_cache_pruning(cache_time_based):
    cache = cache_time_based
    """ Test that when adding more cache records than the threshold then
    some records get pruned, so the size of the cache stays at or below the threshold + 1
    """
    # for the Redis cache there is no pruning implemented, so we skip this test
    if isinstance(cache, Redis)\
            or isinstance(cache, RedisSentinel):
        pytest.skip("There is no pruning for Redis")

    # for the Memcache cache there is no pruning implemented, so we skip this test
    if isinstance(cache, MemcachedCache)\
            or isinstance(cache, SASLMemcachedCache)\
            or isinstance(cache, SpreadSASLMemcachedCache):
        pytest.skip("There is no pruning for Memcached")

    for i in range(CACHE_THRESHOLD + 5):
        cache.set(f"hi-{i}", "hello")

    # the size of the cache is at or below the cache threshold:
    num_cached = 0
    for i in range(CACHE_THRESHOLD + 5):
        num_cached += 1 if cache.has(f"hi-{i}") else 0

    assert num_cached <= CACHE_THRESHOLD + 1



""" The different options available """


class CacheEvictionStrategy:
    """ All implemented cache eviction strategies """
    time_based = 'time-based'
    rest_based = 'rest-based'
    rest_and_time_based = 'rest-and-time-based'


class HttpMethods:
    """ All HTTP methods available """
    GET = 'GET'
    POST = 'POST'
    PATCH = 'PATCH'
    PUT = 'PUT'
    DELETE = 'DELETE'



# -*- coding: utf-8 -*-
"""
    flask_caching.backends.memcache
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    The memcache caching backend.

    :copyright: (c) 2019 by Zoltan Fedor.
    :license: MIT, see LICENSE for more details.

    :copyright: (c) 2018 by Peter Justin.
    :copyright: (c) 2010 by Thadeus Burgess.
    :license: BSD, see LICENSE for more details.
"""
import re
from time import time

from falcon_caching.backends.base import BaseCache, iteritems_wrapper

try:
    import cPickle as pickle
except ImportError:  # pragma: no cover
    import pickle


_test_memcached_key = re.compile(r"[^\x00-\x21\xff]{1,250}$").match


class MemcachedCache(BaseCache):

    """A cache that uses memcached as backend.

    The first argument can either be an object that resembles the API of a
    :class:`memcache.Client` or a tuple/list of server addresses. In the
    event that a tuple/list is passed, Werkzeug tries to import the best
    available memcache library.

    This cache looks into the following packages/modules to find bindings for
    memcached:

        - ``pylibmc``
        - ``google.appengine.api.memcached``
        - ``memcached``
        - ``libmc``

    Implementation notes:  This cache backend works around some limitations in
    memcached to simplify the interface.  For example unicode keys are encoded
    to utf-8 on the fly.  Methods such as :meth:`~BaseCache.get_dict` return
    the keys in the same format as passed.  Furthermore all get methods
    silently ignore key errors to not cause problems when untrusted user data
    is passed to the get methods which is often the case in web applications.

    :param servers: a list or tuple of server addresses or alternatively
                    a :class:`memcache.Client` or a compatible client.
    :param default_timeout: the default timeout that is used if no timeout is
                            specified on :meth:`~BaseCache.set`. A timeout of
                            0 indicates that the cache never expires.
    :param key_prefix: a prefix that is added before all keys.  This makes it
                       possible to use the same memcached server for different
                       applications.  Keep in mind that
                       :meth:`~BaseCache.clear` will also clear keys with a
                       different prefix.
    """

    def __init__(self, servers=None, default_timeout=300, key_prefix=None):
        super(MemcachedCache, self).__init__(default_timeout)
        if servers is None or isinstance(servers, (list, tuple)):
            if servers is None:
                servers = ["127.0.0.1:11211"]
            self._client = self.import_preferred_memcache_lib(servers)
            if self._client is None:
                raise RuntimeError("no memcache module found")
        else:
            # NOTE: servers is actually an already initialized memcache
            # client.
            self._client = servers

        self.key_prefix = key_prefix or None

    def _normalize_key(self, key):
        key = str(key)
        if self.key_prefix:
            key = self.key_prefix + key
        return key

    def _normalize_timeout(self, timeout):
        timeout = BaseCache._normalize_timeout(self, timeout)
        if timeout > 0:
            # NOTE: pylibmc expect the timeout as delta time up to
            # 2592000 seconds (30 days)
            if not hasattr(self, 'mc_library'):
                try:
                    import pylibmc
                except ImportError:
                    self.mc_library = None
                else:
                    self.mc_library = 'pylibmc'

            if self.mc_library != 'pylibmc':
                timeout = int(time()) + timeout
            elif timeout > 2592000:
                timeout = 0

        return timeout

    def get(self, key):
        key = self._normalize_key(key)
        # memcached doesn't support keys longer than that.  Because often
        # checks for so long keys can occur because it's tested from user
        # submitted data etc we fail silently for getting.
        if _test_memcached_key(key):
            return self._client.get(key)

    def get_dict(self, *keys):
        key_mapping = {}
        have_encoded_keys = False
        for key in keys:
            encoded_key = self._normalize_key(key)
            if not isinstance(key, str):
                have_encoded_keys = True
            if _test_memcached_key(key):
                key_mapping[encoded_key] = key
        _keys = list(key_mapping)
        d = rv = self._client.get_multi(_keys)
        if have_encoded_keys or self.key_prefix:
            rv = {}
            for key, value in d.items():
                rv[key_mapping[key]] = value
        if len(rv) < len(keys):
            for key in keys:
                if key not in rv:
                    rv[key] = None
        return rv

    def add(self, key, value, timeout=None):
        key = self._normalize_key(key)
        timeout = self._normalize_timeout(timeout)
        return self._client.add(key, value, timeout)

    def set(self, key, value, timeout=None):
        key = self._normalize_key(key)
        timeout = self._normalize_timeout(timeout)
        return self._client.set(key, value, timeout)

    def get_many(self, *keys):
        d = self.get_dict(*keys)
        return [d[key] for key in keys]

    def set_many(self, mapping, timeout=None):
        new_mapping = {}
        for key, value in iteritems_wrapper(mapping):
            key = self._normalize_key(key)
            new_mapping[key] = value

        timeout = self._normalize_timeout(timeout)
        failed_keys = self._client.set_multi(new_mapping, timeout)
        return not failed_keys

    def delete(self, key):
        key = self._normalize_key(key)
        if _test_memcached_key(key):
            return self._client.delete(key)

    def delete_many(self, *keys):
        new_keys = []
        for key in keys:
            key = self._normalize_key(key)
            if _test_memcached_key(key):
                new_keys.append(key)
        return self._client.delete_multi(new_keys)

    def has(self, key):
        key = self._normalize_key(key)
        if _test_memcached_key(key):
            try:
                return self._client.append(key, "")
            except AttributeError:
                # GAEMemecache has no 'append' function
                return True if self._client.get(key) is not None else False
        return False

    def clear(self):
        return self._client.flush_all()

    def inc(self, key, delta=1):
        key = self._normalize_key(key)
        return self._client.incr(key, delta)

    def dec(self, key, delta=1):
        key = self._normalize_key(key)
        return self._client.decr(key, delta)

    def import_preferred_memcache_lib(self, servers):
        """Returns an initialized memcache client.  Used by the constructor."""
        try:
            import pylibmc
        except ImportError:
            pass
        else:
            self.mc_library = 'pylibmc'
            return pylibmc.Client(servers)

        try:
            from google.appengine.api import memcache
        except ImportError:
            pass
        else:
            self.mc_library = 'google.appengine.api'
            return memcache.Client()

        try:
            import memcache
        except ImportError:
            pass
        else:
            self.mc_library = 'memcache'
            return memcache.Client(servers)

        try:
            import libmc
        except ImportError:
            pass
        else:
            self.mc_library = 'libmc'
            return libmc.Client(servers)


class SASLMemcachedCache(MemcachedCache):
    def __init__(
        self,
        servers=None,
        default_timeout=300,
        key_prefix=None,
        username=None,
        password=None,
        **kwargs
    ):
        super(SASLMemcachedCache, self).__init__(default_timeout)

        if servers is None:
            servers = ["127.0.0.1:11211"]

        import pylibmc

        self._client = pylibmc.Client(
            servers, username=username, password=password, binary=True, **kwargs
        )

        self.key_prefix = key_prefix


class SpreadSASLMemcachedCache(SASLMemcachedCache):
    """Simple Subclass of SASLMemcached client that will spread the value
    across multiple keys if they are bigger than a given treshold.

    Spreading requires using pickle to store the value, which can significantly
    impact the performance.
    """

    def __init__(self, *args, **kwargs):
        """
        Kwargs:
            chunksize (int): max length of a pickled object that can fit in
                memcached (memcache has an upper limit of 1MB for values,
                default: 1048448)
        """
        self.chunksize = kwargs.get("chunksize", 1048448)
        self.maxchunk = kwargs.get("maxchunk", 32)
        super(SpreadSASLMemcachedCache, self).__init__(*args, **kwargs)

    def delete(self, key):
        for skey in self._genkeys(key):
            super(SpreadSASLMemcachedCache, self).delete(skey)

    def delete_many(self, *keys):
        for key in keys:
            self.delete(key)

    def set(self, key, value, timeout=None, chunk=True):
        """Set a value in cache, potentially spreading it across multiple key.

        :param key: The cache key.
        :param value: The value to cache.
        :param timeout: The timeout after which the cache will be invalidated.
        :param chunk: If set to `False`, then spreading across multiple keys
                      is disabled. This can be faster, but it will fail if
                      the value is bigger than the chunks. It requires you
                      to get back the object by specifying that it is not
                      spread.
        """
        if chunk:
            return self._set(key, value, timeout=timeout)
        else:
            return super(SpreadSASLMemcachedCache, self).set(
                key, value, timeout=timeout
            )

    def _set(self, key, value, timeout=None):
        # pickling/unpickling add an overhead,
        # I didn't found a good way to avoid pickling/unpickling if
        # key is smaller than chunksize, because in case or <werkzeug.requests>
        # getting the length consume the data iterator.
        serialized = pickle.dumps(value, 2)
        values = {}
        len_ser = len(serialized)
        chks = range(0, len_ser, self.chunksize)

        if len(chks) > self.maxchunk:
            raise ValueError(
                "Cannot store value in less than %s keys" % self.maxchunk
            )

        for i in chks:
            values["%s.%s" % (key, i // self.chunksize)] = serialized[
                i: i + self.chunksize
            ]

        super(SpreadSASLMemcachedCache, self).set_many(values, timeout)

    def get(self, key, chunk=True):
        """Get a cached value.

        :param chunk: If set to ``False``, it will return a cached value
                      that is spread across multiple keys.
        """
        if chunk:
            return self._get(key)
        else:
            return super(SpreadSASLMemcachedCache, self).get(key)

    def _genkeys(self, key):
        return ["%s.%s" % (key, i) for i in range(self.maxchunk)]

    def _get(self, key):
        to_get = ["%s.%s" % (key, i) for i in range(self.maxchunk)]
        result = super(SpreadSASLMemcachedCache, self).get_many(*to_get)
        serialized = b"".join(v for v in result if v is not None)

        if not serialized:
            return None

        return pickle.loads(serialized)

    def has(self, key):
        return super().has("%s.%s" % (key, 0))




from falcon_caching.backends.filesystem import FileSystemCache
from falcon_caching.backends.memcache import (
    MemcachedCache,
    SASLMemcachedCache,
    SpreadSASLMemcachedCache
)
from falcon_caching.backends.null import NullCache
from falcon_caching.backends.redis import Redis, RedisSentinel
from falcon_caching.backends.simple import SimpleCache

try:
    from falcon_caching.backends.uwsgicache import UWSGICache

    has_UWSGICache = True
except ImportError:
    has_UWSGICache = False

__all__ = (
    "null",
    "simple",
    "filesystem",
    "redis",
    "redissentinel",
    "uwsgi",
    "memcached",
    "gaememcached",
    "saslmemcached",
    "spreadsaslmemcached",
)


def null(config, args, kwargs):
    return NullCache()


def simple(config, args, kwargs):
    kwargs.update(
        dict(
            threshold=config["CACHE_THRESHOLD"],
            ignore_errors=config["CACHE_IGNORE_ERRORS"],
        )
    )
    return SimpleCache(*args, **kwargs)


def filesystem(config, args, kwargs):
    args.insert(0, config["CACHE_DIR"])
    kwargs.update(
        dict(
            threshold=config["CACHE_THRESHOLD"],
            ignore_errors=config["CACHE_IGNORE_ERRORS"],
        )
    )
    return FileSystemCache(*args, **kwargs)


def redis(config, args, kwargs):
    try:
        from redis import from_url as redis_from_url
    except ImportError:
        raise RuntimeError("no redis module found")

    kwargs.update(
        dict(
            host=config.get("CACHE_REDIS_HOST", "localhost"),
            port=config.get("CACHE_REDIS_PORT", 6379),
        )
    )
    password = config.get("CACHE_REDIS_PASSWORD")
    if password:
        kwargs["password"] = password

    key_prefix = config.get("CACHE_KEY_PREFIX")
    if key_prefix:
        kwargs["key_prefix"] = key_prefix

    db_number = config.get("CACHE_REDIS_DB")
    if db_number:
        kwargs["db"] = db_number

    redis_url = config.get("CACHE_REDIS_URL")
    if redis_url:
        kwargs["host"] = redis_from_url(redis_url, db=kwargs.pop("db", None))

    return Redis(*args, **kwargs)


def redissentinel(config, args, kwargs):
    kwargs.update(
        dict(
            sentinels=config.get(
                "CACHE_REDIS_SENTINELS", [("127.0.0.1", 26379)]
            ),
            master=config.get("CACHE_REDIS_SENTINEL_MASTER", "mymaster"),
            password=config.get("CACHE_REDIS_PASSWORD", None),
            sentinel_password=config.get("CACHE_REDIS_SENTINEL_PASSWORD", None),
            key_prefix=config.get("CACHE_KEY_PREFIX", None),
            db=config.get("CACHE_REDIS_DB", 0),
        )
    )

    return RedisSentinel(*args, **kwargs)


def uwsgi(config, args, kwargs):
    if not has_UWSGICache:
        raise NotImplementedError(
            "UWSGICache backend is not available, "
            "you should upgrade werkzeug module."
        )
    # The name of the caching instance to connect to, for
    # example: mycache@localhost:3031, defaults to an empty string, which
    # means uWSGI will cache in the local instance. If the cache is in the
    # same instance as the werkzeug app, you only have to provide the name of
    # the cache.
    uwsgi_cache_name = config.get("CACHE_UWSGI_NAME", "")
    kwargs.update(dict(cache=uwsgi_cache_name))
    return UWSGICache(*args, **kwargs)


def memcached(config, args, kwargs):
    args.append(config["CACHE_MEMCACHED_SERVERS"])
    kwargs.update(dict(key_prefix=config["CACHE_KEY_PREFIX"]))
    return MemcachedCache(*args, **kwargs)


def gaememcached(config, args, kwargs):
    return memcached(config, args, kwargs)


def saslmemcached(config, args, kwargs):
    args.append(config["CACHE_MEMCACHED_SERVERS"])
    kwargs.update(
        dict(
            username=config["CACHE_MEMCACHED_USERNAME"],
            password=config["CACHE_MEMCACHED_PASSWORD"],
            key_prefix=config["CACHE_KEY_PREFIX"],
        )
    )
    return SASLMemcachedCache(*args, **kwargs)


def spreadsaslmemcached(config, args, kwargs):
    args.append(config["CACHE_MEMCACHED_SERVERS"])
    kwargs.update(
        dict(
            username=config.get("CACHE_MEMCACHED_USERNAME"),
            password=config.get("CACHE_MEMCACHED_PASSWORD"),
            key_prefix=config.get("CACHE_KEY_PREFIX"),
        )
    )

    return SpreadSASLMemcachedCache(*args, **kwargs)



from falcon import HTTP_200
from falcon_caching.options import CacheEvictionStrategy, HttpMethods
from typing import TYPE_CHECKING, Any, Dict

if TYPE_CHECKING:
    from falcon_caching.cache import Cache


class Middleware:
    """ It integrates a cache object with Falcon by turning it into
    a Falcon Middleware
    """

    def __init__(self, cache: 'Cache', config: Dict[str, Any]) -> None:
        self.cache = cache
        self.cache_config = config

    def process_resource(self, req, resp, resource, params):
        """ Determine if the given request is marked for caching and if yes,
        then look it up in the cache and if found, then return the cached value
        """

        # Step 1: for 'rest-based' and 'rest&time-based' eviction strategies the
        # POST/PATCH/PUT/DELETE calls are never cached, they should never be
        # loaded from cache as they must always execute,
        # so for those we don't need to try to search the cache
        if self.cache_config['CACHE_EVICTION_STRATEGY'] in [CacheEvictionStrategy.rest_based,
                                                            CacheEvictionStrategy.rest_and_time_based] \
            and req.method.upper() in [HttpMethods.POST,
                                       HttpMethods.PATCH,
                                       HttpMethods.PUT,
                                       HttpMethods.DELETE]:
            return

        # Step 2: determine whether the given responder has caching setup
        # and if not then short-circuit to save on the lookup of request in the cache
        # as anyhow this request was not marked to be cached

        # find out which responder ("on_..." method) is going to be used to process this request
        responder = None
        for _method in dir(resource):
            if _method.startswith('on_') and _method[3:].upper() == req.method.upper():
                responder = _method
                break

        if responder:
            # get the name of the responder wrapper, which for cached objects is 'cache_wrap'
            # see the "Cache.cache" decorator in cache.py
            responder_wrapper_name = getattr(getattr(resource, responder), '__name__')

            if responder_wrapper_name != 'cache_wrap':
                # no caching was requested on this responder
                return

        # Step 3: look up the record in the cache
        key = self.generate_cache_key(req)
        data = self.cache.get(key)

        # "or self.cache.has(key)" was required, because 'data' can be None for one of two reasons:
        #   (1) - the given key wasn't cached yet
        #   (2) - the given key is cached, but the endpoint has returned None in its body
        # which is why if it is None then we also need to check for (2)
        if data or self.cache.has(key):
            resp.body = data
            resp.status = HTTP_200
            req.context['cached'] = True

            # Short-circuit any further processing to skip any remaining
            # 'process_request' and 'process_resource' methods, as well as
            # the 'responder' method that the request would have been routed to.
            # However, any 'process_response' middleware methods will still be called.
            resp.complete = True

    def process_response(self, req, resp, resource, req_succeeded):
        """ Cache the response if this request qualifies and has not been cached yet
        or for rest-based and rest-and-time-based evict the record from the cache if
        the request method is POST/PATCH/PUT or DELETE """

        # Step 1: for 'rest-based' and 'rest&time-based' eviction strategies the
        # POST/PATCH/PUT/DELETE calls are never cached and even more they
        # invalidate the record cached by the GET method
        if self.cache_config['CACHE_EVICTION_STRATEGY'] in [CacheEvictionStrategy.rest_based,
                                                            CacheEvictionStrategy.rest_and_time_based] \
            and req.method.upper() in [HttpMethods.POST,
                                       HttpMethods.PATCH,
                                       HttpMethods.PUT,
                                       HttpMethods.DELETE]:
            # get the cache key created by the GET method (assuming there was one)
            key = self.generate_cache_key(req, method='GET')
            self.cache.delete(key)
            return

        # Step 2: if it is marked to be cached, but has not yet been cached
        # then we cache it
        if 'cache' in req.context and req.context['cache'] \
                and ('cached' not in req.context or not req.context['cached']):
            key = self.generate_cache_key(req)
            value = resp.body

            # for the REST-based strategy there is no timeout, the cached record never expires
            if self.cache_config['CACHE_EVICTION_STRATEGY'] in [CacheEvictionStrategy.rest_based]:
                # timeout 0 - never expires
                timeout = 0
            else:
                # for the time-based and rest-and-time-based eviction strategy the
                # cached record expires
                timeout = req.context.get('cache_timeout', 600)

            self.cache.set(key, value, timeout=timeout)

    @staticmethod
    def generate_cache_key(req, method: str = None) -> str:
        """ Generate the cache key from the request using the path, method and request body """

        # Get the body of the request to be used in the key.
        # If Content-Length happens to be 0, or the header is
        # missing altogether, reading it as below will not block.
        # see https://falcon.readthedocs.io/en/stable/api/request_and_response.html#falcon.Request.stream
        request_body = req.stream.read(req.content_length or 0)
        request_body = request_body.decode() if request_body else ''

        path = req.path
        if path.endswith('/'):
            path = path[:-1]

        if not method:
            method = req.method

        return f'{path}:{method.upper()}:{request_body}'



""" Functional tests to test an example app with the different
cache backends and eviction strategies
"""
import pytest
import time

from falcon_caching.backends.memcache import (
    MemcachedCache,
    SASLMemcachedCache,
    SpreadSASLMemcachedCache
)
from falcon_caching.backends.redis import Redis, RedisSentinel
from falcon_caching.backends.filesystem import FileSystemCache
from falcon_caching.backends.simple import SimpleCache
from tests.conftest import CACHE_EXPIRES, CACHE_BUSTING_METHODS
from tests.utils import get_cache, get_cache_class, get_cache_eviction_strategy,\
    delete_from_cache


def test_app_basic(client):
    # delete any existing cache records for this endpoint to ensure
    # that we start from scratch
    delete_from_cache(app=client.app, path='/randrange_cached', method="GET")
    result1 = client.simulate_get('/randrange_cached')
    assert 100000 >= result1.json['num'] >= 0


def test_onget_cache(client):
    """ Test a properly cached GET request
    """
    # delete any existing cache records for this endpoint to ensure
    # that we start from scratch
    delete_from_cache(app=client.app, path='/randrange_cached', method="GET")

    result1 = client.simulate_get('/randrange_cached')
    # if get_cache_class(client.app) in [SimpleCache, FileSystemCache]:
    #      pytest.skip("Skipping SimpleCache")

    result2 = client.simulate_get('/randrange_cached')

    assert result1.json['num'] == result2.json['num']


@pytest.mark.parametrize("decorator_type", [
    # testing the method-based decorators, eg calling the
    # /randrange_cached_expires endpoint
    ("method-decorated"),
    # testing the class-based decorators, eg calling the
    # /randrange_class_cached_expires endpoint
    ("class-decorated")
])
def test_cache_expires(client, decorator_type):
    """ Test expiring cache with the different eviction strategies both
    for method decorators and class decorators
    """
    if decorator_type == 'method-decorated':
        path = "/randrange_cached_expires"
    elif decorator_type == 'class-decorated':
        path = "/randrange_class_cached_expires"

    # delete any existing cache records for this endpoint to ensure
    # that we start from scratch
    delete_from_cache(app=client.app, path=path, method="GET")

    result1 = client.simulate_get(path)
    result2 = client.simulate_get(path)
    assert result1.json['num'] == result2.json['num']

    # if we wait long enough the cache expires so a new request will bring
    # a new value
    time.sleep(CACHE_EXPIRES)
    result3 = client.simulate_get(path)

    # if the eviction strategy is time-based or rest-and-time-based
    # then the cache should have already expired, so the request
    # should have gotten executed again
    eviction_strategy = get_cache_eviction_strategy(client.app)
    if eviction_strategy in ['time-based', 'rest-and-time-based']:
        assert result2.json['num'] != result3.json['num']
    elif eviction_strategy in ['rest-based']:
        assert result2.json['num'] == result3.json['num']

        # a POST command should evict the cache for rest-based
        client.simulate_post(path)

        result4 = client.simulate_get(path)
        assert result3.json['num'] != result4.json['num']


def test_rest_action_expires_cache(client):
    """ Test how cache records expires after different method requests
    with the different eviction strategies
    """
    # delete any existing cache records for this endpoint to ensure
    # that we start from scratch
    delete_from_cache(app=client.app, path="/randrange_cached", method="GET")
    for method in CACHE_BUSTING_METHODS:
        delete_from_cache(app=client.app, path="/randrange_cached", method=method)

    # normal get should chache for all eviction strategies
    result1 = client.simulate_get('/randrange_cached')
    result2 = client.simulate_get('/randrange_cached')
    assert result1.json['num'] == result2.json['num']

    eviction_strategy = get_cache_eviction_strategy(client.app)

    # if the eviction strategy is time-based then no request with
    # other method for the same endpoint should change that
    # but for the rest-based and rest-and-time-based it does change it
    if eviction_strategy == 'time-based':
        for method in CACHE_BUSTING_METHODS:
            result3 = client.simulate_get('/randrange_cached')

            # make the simulate_post(), simulate_put(), etc call:
            getattr(client, f"simulate_{method.lower()}")('/randrange_cached')

            result4 = client.simulate_get('/randrange_cached')
            assert result3.json['num'] == result4.json['num']

    elif eviction_strategy in ['rest-based', 'rest-and-time-based']:
        for method in CACHE_BUSTING_METHODS:
            result3 = client.simulate_get('/randrange_cached')

            # make the simulate_post(), simulate_put(), etc call:
            getattr(client, f"simulate_{method.lower()}")('/randrange_cached')

            result4 = client.simulate_get('/randrange_cached')
            assert result3.json['num'] != result4.json['num']


def test_explicit_caching(caches):
    """ Testing of explicit caching and retrieving of records
    """
    # it is sufficient to test it with one type of cache
    cache = caches['time-based']

    # delete the 'foo21' record to be sure that it is not left in the cache
    cache.delete_many("foo21", "foo22", "foo23", "foo24", "foo25")

    assert cache.has("foo21") is False

    cache.set("foo21", "bar")
    assert cache.has("foo21") is True

    # add() does not overwrite existing keys
    cache.add("foo21", "bar2")
    assert cache.get("foo21") == 'bar'

    # for SpreadSASLMemcachedCache the add() doesn't work
    if cache.cache.__class__ not in [SpreadSASLMemcachedCache]:
        cache.add("foo22", "bar2")
    else:
        cache.set("foo22", "bar2")
    assert cache.get("foo22") == 'bar2'

    # for SpreadSASLMemcachedCache the add() doesn't work
    if cache.cache.__class__ not in [SpreadSASLMemcachedCache]:
        assert cache.get_many("foo21", "foo22") == ["bar", "bar2"]
        assert cache.get_dict("foo21", "foo22") == {"foo21": "bar", "foo22": "bar2"}

    cache.set_many({"foo23": "bar3", "foo24": "bar4"})
    assert cache.get_many("foo23", "foo24") == ["bar3", "bar4"]

    # for SpreadSASLMemcachedCache the add() doesn't work
    if cache.cache.__class__ not in [SpreadSASLMemcachedCache]:
        cache.delete_many("foo21", "foo22")
        assert cache.has("foo21") is False
        assert cache.has("foo22") is False

        assert cache.has("foo23") is True
        assert cache.has("foo24") is True
        cache.clear()
        assert cache.has("foo23") is False
        assert cache.has("foo24") is False

    # these only work for Redis
    if cache.cache.__class__ in [Redis, RedisSentinel]:
        cache.set("foo25", 1)
        cache.inc("foo25")
        assert cache.get("foo25") == 2

        cache.dec("foo25")
        assert cache.get("foo25") == 1



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['falcon_caching', 'falcon_caching.backends']

package_data = \
{'': ['*']}

install_requires = \
['falcon >=1.0.0']

extras_require = \
{'test': ['flit',
          'pylibmc',
          'pytest >=4.0.0',
          'pytest-cov',
          'pytest-pep8',
          'pytest-xdist',
          'pytest-xprocess',
          'redis',
          'tox',
          'tox-pipenv-install']}

setup(name='falcon_caching',
      version='0.1.0',
      description='Falcon-Caching - a caching module for the Falcon web framework',
      author='Zoltan Fedor',
      author_email='zoltan.1.fedor@gmail.com',
      url='https://github.com/zoltan-fedor/falcon-caching',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      python_requires='>=3.6',
     )
